#pragma once

namespace tf {

// ----------------------------------------------------------------------------
// forward declarations
// ----------------------------------------------------------------------------

// taskflow
class Node;
class Graph;
class FlowBuilder;
class Subflow;
class Task;
class TaskView;
class Taskflow;
class Topology;
class Executor;

// cudaflow
class cudaNode;
class cudaGraph;
class cudaTask;
class cudaFlow;



}  // end of namespace tf -----------------------------------------------------




